\documentclass[a4paper]{article}

\usepackage[latin1]{inputenc}
\usepackage[portuges]{babel}
\usepackage{amsmath}

%include polycode.fmt

\title{Trabalho 1: Compilador para \texttt{MSP}}
\author{Mtodos de Programao 1}

\begin{document}

\maketitle

%if False
\begin{code}
import Data.Map hiding (map)
import Control.Monad.State  
\end{code}
%endif

\section{Introduo}
\label{sec:introducao}

O objectivo deste trabalho  desenvolver um compilador para uma
pequena linguagem de programao designada \texttt{Prog}. Nesta
linguagem s existem duas instrues:
%format lett = "\mathbf{let}"
%format printt = "\mathbf{print}"
\begin{itemize}
\item |lett x = e|, onde |x|  uma varivel e |e| uma expresso
  aritmtica sobre inteiros que lhe  atribuda.
\item |printt e|, que imprime a expresso aritmtica |e|. 
\end{itemize}
As instrues podem ser sequenciadas com ``;'' para compor programas
mais sofisticados, como por exemplo:
\begin{center}
  |lett x = 2; printt (-3) * (4 + x)|
\end{center}

Os seguintes tipos Haskell vo ser usado representar estes programas.
\begin{code}
data Prog  o = Print (Exp o) | Let String (Exp o) | Seq (Prog o) (Prog o)
data Exp   o = Const Int | Var String | Op o [Exp o]  
\end{code}
Uma expresso aritmtica pode ser uma constante, uma varivel ou a
aplicao de um operador a uma lista de argumentos. Ambos os tipos so
parametrizados pelo tipo dos operadores aritmticos, por forma a
tornar o compilador independente do conjunto concreto de operadores a
usar. Por exemplo, se apenas pretendermos ter soma, produto e
simtrico, podemos declarar o seguinte tipo para os operadores.
\begin{code}
data Ops = Add | Mul | Sim  
\end{code}

O pequeno programa acima apresentado poderia agora ser representado
pelo seguinte valor:
\begin{code}
prog :: Prog Ops
prog = Seq  (Let "x" (Const 2)) 
            (Print (Op Mul [  Op Sim [Const 3],
                              Op Add [Const 4, Var "x"]]))
\end{code}

 possvel definir uma funo para executar directamente programas
\texttt{Prog}. Para tal, vamos comear por definir uma funo para
avaliar o valor das expresses aritmticas. Como as expresses podem
conter variveis precisamos de um dicionrio que determine o seu
valor. Este dicionrio  uma correspondncia entre variveis e valores
e pode ser implementado com o tipo |Map| da biblioteca |Data.Map|:
\begin{code}
type VarDict = Map String Int  
\end{code}

Tambm  necessrio saber qual a funo Haskell que corresponde a cada
um dos nossos operadores. Para tal vamos definir uma classe |Opt| que
agrupa todos os operadores.
\begin{code}
class Opt o where
    arity  :: o -> Int
    func   :: o -> ([Int] -> Int)  
\end{code}
A primeira funo desta classe diz-nos qual a aridade de um
operador. A segunda devolve uma funo que pode ser usada para o
executar. A instncia desta classe para o nosso tipo |Ops|  a
seguinte:
\begin{code}
instance Opt Ops where
    arity  Add = 2
    arity  Mul = 2
    arity  Sim = 1
    func   Add = \[x,y]  -> x+y
    func   Mul = \[x,y]  -> x*y
    func   Sim = \[x]    -> -x   
\end{code}

 agora possvel definir a funo |eval| que avalia o valor de uma
expresso.
\begin{code}
eval :: Opt o => VarDict -> Exp o -> Int
eval vd (Const x)  = x
eval vd (Var v)    = vd ! v
eval vd (Op o l)   = func o (map (eval vd) l)  
\end{code}

A funo que executa um programa em \texttt{Prog} tem que actualizar um
dicionrio sempre que encontra uma instruo |lett|. Este dicionrio 
depois passado  funo |eval| sempre que uma expresso 
encontrada. A forma mais simples de implementar este algoritmo passa
pela utilizao do \emph{monad} estado da biblioteca
|Control.Monad.State|. No entanto, como tambm precisamos do \emph{monad}
|IO| (para executar a instruo |printt|) temos que combinar os dois
recorrendo ao transformador de \emph{monads} |StateT|.
\begin{code}
execprog :: Opt o => Prog o -> IO ()
execprog p = evalStateT (aux p) empty
    where aux :: Opt o => Prog o -> StateT VarDict IO ()  
          aux (Print e)  = do  vd <- get
			       lift (print (eval vd e))
	  aux (Let v e)  = do  vd <- get
                               put (insert v (eval vd e) vd)
	  aux (Seq l r)  = do  aux l
			       aux r  
\end{code}

Podemos agora usar a funo |execprog| para executar o nosso programa.
% Cuidado com o > no incio da linha
\begin{spec}
 > execprog prog
-18  
\end{spec}

\section{A mquina abstracta \texttt{MSP}}
\label{sec:maqu-abstr-msp}

Em vez de executar directamente estes programas pretende-se agora
desenvolver um compilador para a mquina abstracta \texttt{MSP}
(\emph{Mais Simples Possvel}). Ser conveniente conhecer bem a
\texttt{MSP}, pelo que se recomenda a leitura do respectivo manual,
disponibilizado no \emph{kit} do trabalho.

A \texttt{MSP} possui duas reas de memria: a \emph{heap} que serve para
armazenar valores usados pelo programa (no caso da nossa linguagem as
variveis) e a \emph{stack} que  uma memria auxiliar  execuo das
instrues (e que funciona segundo a filosofia habitual deste tipo de
dados). Para simplificar, vamos assumir que ambas as reas de memria
so listas de inteiros: a \emph{heap}  uma lista de tamanho fixo igual ao
tamanho desejado e a \emph{stack} tem tamanho varivel, sendo as inseres e
remoes feitas na cabea da lista. Podemos representar a memria pelo
seguinte tipo de dados:
\begin{code}
data Mem = Mem {stack :: [Int], heap :: [Int]}  
\end{code}
Assumindo que a \emph{heap} tem 1024 posies de memria, o estado inicial da
memria  representado pelo seguinte valor.
\begin{code}
emptymem :: Mem
emptymem = Mem {stack = [], heap = replicate 1024 undefined}  
\end{code}

O conjunto de instrues da mquina \texttt{MSP} que nos interessa para este
trabalho  bastante restricto:
\begin{description}
\item[PUSH $x$] Acrescenta o inteiro $x$ ao topo da \emph{stack}.
\item[LOAD] Retira um endereo do topo da \emph{stack} e substitui-o pelo o
  contedo desse endereo na \emph{heap}.
\item[STORE] Retira um valor e um endereo do topo da \emph{stack} e armazena
  esse valor no respectivo endereo da \emph{heap}.
\item[IN] L um inteiro do teclado e armazena-o no topo da \emph{stack}.
\item[OUT] Retira um valor do topo da \emph{stack} e imprime-o no monitor.
\item[OP $o$] Dado um operador $o$, retira os seus argumentos do topo
  da \emph{stack} e substitui-os pelo resultado da operao respectiva.
\end{description}
Estas instrues podem ser representadas pelo seguinte tipo de dados:
\begin{code}
data Instr o = PUSH Int | LOAD | STORE | IN | OUT | OP o   
\end{code}
Um programa em \texttt{MSP}  simplesmente uma sequncia destas instrues:
\begin{code}
type MSP o = [Instr o]  
\end{code}
Por exemplo, o seguinte programa imprime o resultado de somar 2 com 3
no monitor.
\begin{code}
msp :: MSP Ops
msp = [PUSH 3, PUSH 2, OP Add, OUT]  
\end{code}

 possvel compilar programas escritos em \texttt{Prog} para \texttt{MSP}. Durante a
compilao  necessrio atribuir um endereo na \emph{heap} para cada
varivel declarada no programa. Supondo que a varivel $x$ foi alocada
ao endereo 0, o programa exemplo acima apresentado seria compilado no
seguinte programa \texttt{MSP}.
\begin{code}
mspprog :: MSP Ops
mspprog = [  PUSH 0,PUSH 2,STORE,PUSH 3,OP Sim, 
             PUSH 4,PUSH 0,LOAD,OP Add,OP Mul,OUT ]
\end{code}
As 3 primeiras instrues deste programa servem para armazenar o valor
2 no endereo 0 atribudo  varivel $x$. Como  habitual nas mquinas
de \emph{stack}, para realizar uma dada operao  necessrio fazer antes o
\emph{push} dos seus argumentos. Este facto implica que a compilao de uma
expresso seja feita atravs de uma travessia \emph{postorder} da sua
rvore sinttica.

\section{Tarefas a realizar}

Neste trabalho devero ser implementadas no mnimo as seguintes
funes:
\begin{itemize}
\item |execmsp :: Opt o => MSP o -> IO ()|\\Esta funo  responsvel
  por executar um programa \texttt{MSP}. Pretende-se que seja usada
  uma combinao dos \emph{monads} |State| e |IO| para o fazer, sendo
  o estado a memria da mquina em cada instante.
\item |compile :: Opt o => Prog o -> MSP o|\\Esta funo deve compilar
  programas escritos em \texttt{Prog} para \texttt{MSP}.
\end{itemize}
Estas funes devem satisfazer o seguinte critrio de correco, que
obriga a que o resultado de executar um programa antes e depois da
compilao seja idntico.
\begin{center}
  |execprog = execmsp . compile|
\end{center}

Para obter nota de Bom ou superior devem implementar algumas
funcionalidades extra, como por exemplo:
\begin{itemize}
\item Implementar instncias das classes |Show| e |Read| para os tipos
  |Prog| e |MSP| e criar programas executveis para compilar um
  programa de \texttt{Prog} para \texttt{MSP} e para executar os programas
  resultantes.
\item Acrescentar funcionalidades de \emph{debug} s funes |compile|
  e |execmsp|. Existem vrios erros de compilao e execuo que podem
  ocorrer: operadores com nmero de argumentos errado, variveis no
  definidas, endereos de memria errados, tentativa de retirar
  valores de uma \emph{stack} vazia, etc. O tratamento de erros deve ser
  feito usando o \emph{monad} |Error|.
\item Estender o conjunto de instrues da mquina virtual
  \texttt{MSP} e da linguagem de programao \texttt{Prog}, por forma
  a obter uma linguagem \texttt{Prog++}. Esta linguagem pode, por
  exemplo, conter mais operadores aritmticos, instrues de leitura
  do teclado, mais tipos de dados, etc. Se desejar, pode alterar a
  classe |Opt| ou mesmo usar uma estratgia diferente para representar
  os operadores.
\end{itemize}

O trabalho deve ser entregue no \emph{site} respectivo at ao dia 5 de
Novembro. Deve ser realizado em grupos de 3 alunos e o relatrio
escrito em \emph{literate} Haskell. Este documento foi gerado a partir
de uma fonte \emph{literate} Haskell usando a ferramenta
\texttt{lhs2tex}:
\begin{center}
  \verb|http://www.informatik.uni-bonn.de/~loeh/lhs2tex/|
\end{center}

Boa Sorte!

\end{document}